using Microsoft.Extensions.Hosting;
using Microsoft.Extensions.Logging;
using System;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace DisplayWorker
{
    public class Worker : BackgroundService
    {
        private readonly ILogger<Worker> _logger;

        public Worker(ILogger<Worker> logger)
        {
            _logger = logger;
            
        }

        Iot.Device.Ssd1351.Ssd1351? display;

        protected override async Task ExecuteAsync(CancellationToken stoppingToken)
        {
            display = CreateDisplay();
            await LoadCalendar();
            await InitDisplayAsync(display);
            Microsoft.Win32.SystemEvents.PowerModeChanged += SystemEvents_PowerModeChanged;
            try
            {
                while (!stoppingToken.IsCancellationRequested)
                {
                    DrawImage(DrawCalendar);
                    await Task.Delay(5000, stoppingToken);
                    DrawImage((g, w, h) =>
                    {
                        g.DrawString(DateTime.Now.ToString("HH:mm"), new System.Drawing.Font("Arial", 24), System.Drawing.Brushes.White, new System.Drawing.PointF(15, h / 2 - 12));
                    });
                    await Task.Delay(1000, stoppingToken);
                }
            }
            catch (OperationCanceledException)
            {
            }
            finally
            {
                display.SetDisplayOff();
                Microsoft.Win32.SystemEvents.PowerModeChanged -= SystemEvents_PowerModeChanged;
            }
        }

        private void SystemEvents_PowerModeChanged(object sender, Microsoft.Win32.PowerModeChangedEventArgs e)
        {
            if (e.Mode == Microsoft.Win32.PowerModes.Suspend)
            {
                display?.SetDisplayOff();
                _logger.LogInformation("Standby detected. Turning off OLED: {time}", DateTimeOffset.Now);
            }
            else if (e.Mode == Microsoft.Win32.PowerModes.Resume)
            {
                display?.SetDisplayOn();
                _logger.LogInformation("Resume detected. Turning on OLED: {time}", DateTimeOffset.Now);
            }
        }

        private void DrawImage(Action<System.Drawing.Graphics, int, int> drawAction)
        {
            using System.Drawing.Bitmap dotnetBM = new(128, 128);
            using var g = System.Drawing.Graphics.FromImage(dotnetBM);
            g.Clear(System.Drawing.Color.Black);
            drawAction(g, 128, 128);
            dotnetBM.RotateFlip(System.Drawing.RotateFlipType.RotateNoneFlipX);
            display?.SendBitmap(dotnetBM);
        }

        #region Display creation

        private static Iot.Device.Ssd1351.Ssd1351 CreateDisplay()
        {
            var devices = Iot.Device.Ft232H.Ft232HDevice.GetFt232H();
            var device = new Iot.Device.Ft232H.Ft232HDevice(devices[0]);
            var driver = device.CreateGpioDriver();            
            var ftSpi = device.CreateSpiDevice(new System.Device.Spi.SpiConnectionSettings(0, 4) { Mode = System.Device.Spi.SpiMode.Mode3, DataBitLength = 8, ClockFrequency = 16_000_000 /* 16MHz */ });
            var controller = new System.Device.Gpio.GpioController(System.Device.Gpio.PinNumberingScheme.Logical, driver);
            Iot.Device.Ssd1351.Ssd1351 display = new(ftSpi, dataCommandPin: 5, resetPin: 6, gpioController: controller);
            return display;
        }

        private static async Task InitDisplayAsync(Iot.Device.Ssd1351.Ssd1351 display)
        {
            await display.ResetDisplayAsync();
            display.Unlock();
            display.MakeAccessible(); // Command A2,B1,B3,BB,BE,C1 accessible if in unlock state
            display.SetDisplayOff(); // Turn on sleep mode
            display.SetDisplayEnhancement(true);
            display.SetDisplayClockDivideRatioOscillatorFrequency(0x00, 0x0F); // 7:4 = Oscillator Frequency, 3:0 = CLK Div Ratio (A[3:0]+1 = 1..16)
            display.SetMultiplexRatio(); // Use all 128 common lines by default....
            display.SetSegmentReMapColorDepth(Iot.Device.Ssd1351.ColorDepth.ColourDepth262K, 
                Iot.Device.Ssd1351.CommonSplit.OddEven, Iot.Device.Ssd1351.Seg0Common.Column0,
                Iot.Device.Ssd1351.ColorSequence.RGB); // 0x74 Color Depth = 64K, Enable COM Split Odd Even, Scan from COM[N-1] to COM0. Where N is the Multiplex ratio., Color sequence is normal: B -> G -> R
            display.SetColumnAddress(); // Columns = 0 -> 127
            display.SetRowAddress(); // Rows = 0 -> 127
            display.SetDisplayStartLine(); // set startline to to 0
            display.SetDisplayOffset(0); // Set vertical scroll by Row to 0-127.
            display.SetGpio(Iot.Device.Ssd1351.GpioMode.Disabled, Iot.Device.Ssd1351.GpioMode.Disabled); // Set all GPIO to Input disabled
            display.SetVDDSource(); // Enable internal VDD regulator
            display.SetPreChargePeriods(2, 3); // Phase 1 period of 5 DCLKS,  Phase 2 period of 3 DCLKS
            display.SetPreChargeVoltageLevel(31);
            display.SetVcomhDeselectLevel(); // 0.82 x VCC
            display.SetNormalDisplay(); // Reset to Normal Display
            display.SetContrastABC(0xC8, 0x80, 0xC8); // Contrast A = 200, B = 128, C = 200
            display.SetMasterContrast(0x0a); // No Change = 15
            display.SetVSL(); // External VSL
            display.Set3rdPreChargePeriod(0x01); // Set Second Pre-charge Period = 1 DCLKS
            display.SetDisplayOn(); //--turn on oled panel
            display.ClearScreen();
        }
        #endregion

        #region Calendar

        private Windows.ApplicationModel.Appointments.AppointmentStore? store;
        private System.Collections.Generic.IList<Windows.ApplicationModel.Appointments.Appointment> events =
            new System.Collections.Generic.List<Windows.ApplicationModel.Appointments.Appointment>();

        private async Task LoadCalendar()
        {
            store = await Windows.ApplicationModel.Appointments.AppointmentManager.RequestStoreAsync(Windows.ApplicationModel.Appointments.AppointmentStoreAccessType.AllCalendarsReadOnly);
            await LoadAppointments();
            store.StoreChanged += (s, e) => LoadAppointments();
        }
        private async Task LoadAppointments()
        {
            if (store is null)
                return;
            var ap = await store.FindAppointmentsAsync(DateTime.Now.Date, TimeSpan.FromDays(90));
            events = ap.Where(a => !a.IsCanceledMeeting).OrderBy(a => a.StartTime).ToList();
        }

        private void DrawCalendar(System.Drawing.Graphics g, int width, int height)
        {
            float y = 0;
            var titleFont = new System.Drawing.Font("Arial", 12);
            var subtitleFont = new System.Drawing.Font("Arial", 10);
            float lineSpacing = 5;
            foreach (var a in events)
            {
                var time = a.StartTime.ToLocalTime();
                if (time + a.Duration < DateTimeOffset.Now)
                {
                    _ = LoadAppointments();
                    continue;
                }
                string timeString = a.AllDay ? time.ToString("d") : time.ToString();
                if (time < DateTimeOffset.Now.Date.AddDays(1)) //Within 24 hours
                {
                    if (a.AllDay) timeString = time.Date <= DateTimeOffset.Now.Date ? "Today" : "Tomorrow";
                    else timeString = time.ToString("HH:mm");
                }
                else if (time < DateTimeOffset.Now.AddDays(7)) // Add day of week
                {
                    if (a.AllDay) timeString = time.ToString("ddd");
                    else timeString = time.ToString("ddd HH:mm");
                }
                if (!a.AllDay && a.Duration.TotalMinutes >= 2) // End time
                {
                    timeString += " - " + time.Add(a.Duration).ToString("HH:mm");
                }
                var color = System.Drawing.Brushes.White;
                if (!a.AllDay && a.StartTime <= DateTimeOffset.UtcNow && a.StartTime + a.Duration > DateTimeOffset.UtcNow)
                    color = System.Drawing.Brushes.Green; // Active meetings are green
                g.DrawString(a.Subject, titleFont, color, new System.Drawing.PointF(0, y));
                y += titleFont.Size + lineSpacing;
                g.DrawString(timeString, subtitleFont, System.Drawing.Brushes.Gray, new System.Drawing.PointF(2, y));
                y += subtitleFont.Size + lineSpacing;
                if (y > height)
                    break;
            }
        }
        #endregion
    }
}